import os
import sys

from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow.metadata.util import sync_local_metadata_to_datastore
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.sidecar import Sidecar
from metaflow.plugins.timeout_decorator import get_run_time_limit_for_task
from .nvcf import NvcfException

from metaflow.metadata import MetaDatum


class NvcfDecorator(StepDecorator):
    name = "nvidia"
    # defaults = {"function_id": "9e5647f2-740f-4101-a129-1c961a075575"}
    defaults = {}
    # "0817006f-018b-4590-b2a5-6cf9d64d9d9a"}
    #

    package_url = None
    package_sha = None

    # Refer https://github.com/Netflix/metaflow/blob/master/docs/lifecycle.png
    # to understand where these functions are invoked in the lifecycle of a
    # Metaflow flow.
    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # Executing NVCF functions requires a non-local datastore.
        if flow_datastore.TYPE not in ("s3", "azure", "gs"):
            raise NvcfException(
                "The *@nvidia* decorator requires --datastore=s3 or --datastore=azure or --datastore=gs at the moment."
            )
        # if self.attributes["function_id"] is None:
        #     raise NvcfException(
        #         "The *@nvidia* decorator requires a function_id. Please reach out to Outerbounds if you are unsure how to get access to one."
        #     )
        # Set internal state.
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

        if any([deco.name == "kubernetes" for deco in decos]):
            raise MetaflowException(
                "Step *{step}* is marked for execution both on Kubernetes and "
                "Nvidia. Please use one or the other.".format(step=step)
            )
        if any([deco.name == "parallel" for deco in decos]):
            raise MetaflowException(
                "Step *{step}* contains a @parallel decorator "
                "with the @nvidia decorator. @parallel is not supported with @nvidia.".format(
                    step=step
                )
            )

        # Set run time limit for the NVCF function.
        self.run_time_limit = get_run_time_limit_for_task(decos)
        if self.run_time_limit < 60:
            raise NvcfException(
                "The timeout for step *{step}* should be at least 60 seconds for "
                "execution with @nvidia.".format(step=step)
            )

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        if retry_count <= max_user_code_retries:
            # after all attempts to run the user code have failed, we don't need
            # to execute on NVCF anymore. We can execute possible fallback
            # code locally.
            cli_args.commands = ["nvidia", "step"]
            cli_args.command_args.append(self.package_sha)
            cli_args.command_args.append(self.package_url)
            cli_args.command_options.update(self.attributes)
            # cli_args.command_options["run-time-limit"] = self.run_time_limit
            cli_args.entrypoint[0] = sys.executable

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        self.metadata = metadata
        self.task_datastore = task_datastore

        # task_pre_step may run locally if fallback is activated for @catch
        # decorator.

        if "NVCF_CONTEXT" in os.environ:
            meta = {}

            meta["nvcf-function-id"] = os.environ.get("NVCF_FUNCTION_ID")
            meta["nvcf-function-version-id"] = os.environ.get(
                "NVCF_FUNCTION_VERSION_ID"
            )
            meta["nvcf-region"] = os.environ.get("NVCF_REGION")
            meta["nvcf-ncaid"] = os.environ.get("NVCF_NCAID")
            meta["nvcf-sub"] = os.environ.get("NVCF_SUB")
            meta["nvcf-instancetype"] = os.environ.get("NVCF_INSTANCETYPE")
            meta["nvcf-reqid"] = os.environ.get("NVCF_REQID")
            meta["nvcf-env"] = os.environ.get("NVCF_ENV")
            meta["nvcf-backend"] = os.environ.get("NVCF_BACKEND")
            meta["nvcf-function-name"] = os.environ.get("NVCF_FUNCTION_NAME")
            meta["nvcf-nspectid"] = os.environ.get("NVCF_NSPECTID")

            entries = [
                MetaDatum(field=k, value=v, type=k, tags=[])
                for k, v in meta.items()
                if v is not None
            ]
            # Register book-keeping metadata for debugging.
            metadata.register_metadata(run_id, step_name, task_id, entries)

            self._save_logs_sidecar = Sidecar("save_logs_periodically")
            self._save_logs_sidecar.start()

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_retries
    ):
        # task_finished may run locally if fallback is activated for @catch
        # decorator.
        if "NVCF_CONTEXT" in os.environ:
            # If `local` metadata is configured, we would need to copy task
            # execution metadata from the NVCF container to user's
            # local file system after the user code has finished execution.
            # This happens via datastore as a communication bridge.
            if hasattr(self, "metadata") and self.metadata.TYPE == "local":
                sync_local_metadata_to_datastore(
                    DATASTORE_LOCAL_DIR, self.task_datastore
                )

        try:
            self._save_logs_sidecar.terminate()
        except:
            # Best effort kill
            pass

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]



import json
import os
import threading
from urllib.parse import urlparse
from urllib.request import HTTPError, Request, URLError, urlopen

from metaflow import util
from metaflow.exception import MetaflowException
from metaflow.mflog import (
    BASH_SAVE_LOGS,
    bash_capture_logs,
    export_mflog_env_vars,
    tail_logs,
    get_log_tailer,
)
import requests
from metaflow.metaflow_config_funcs import init_config


class NvcfException(MetaflowException):
    headline = "Nvidia error"


class NvcfKilledException(MetaflowException):
    headline = "Nvidia job killed"


# Redirect structured logs to $PWD/.logs/
LOGS_DIR = "$PWD/.logs"
STDOUT_FILE = "mflog_stdout"
STDERR_FILE = "mflog_stderr"
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)


class Nvcf(object):
    def __init__(self, metadata, datastore, environment):
        self.metadata = metadata
        self.datastore = datastore
        self.environment = environment

    def launch_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        env={},
    ):
        mflog_expr = export_mflog_env_vars(
            datastore_type=code_package_ds,
            stdout_path=STDOUT_PATH,
            stderr_path=STDERR_PATH,
            **task_spec,
        )
        init_cmds = self.environment.get_package_commands(
            code_package_url, code_package_ds
        )
        init_expr = " && ".join(init_cmds)
        heartbeat_expr = f'python -m metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store "$MAIN_PID" {code_package_ds} nvcf_heartbeats 1>> $MFLOG_STDOUT 2>> $MFLOG_STDERR'
        step_expr = bash_capture_logs(
            " && ".join(
                self.environment.bootstrap_commands(step_name, code_package_ds)
                + [step_cli + " & MAIN_PID=$!; " + heartbeat_expr]
            )
        )

        # construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        cmd_str = "mkdir -p %s && %s && %s && %s; " % (
            LOGS_DIR,
            mflog_expr,
            init_expr,
            step_expr,
        )
        # after the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario.
        cmd_str += "c=$?; %s; exit $c" % BASH_SAVE_LOGS
        cmd_str = (
            '${METAFLOW_INIT_SCRIPT:+eval \\"${METAFLOW_INIT_SCRIPT}\\"} && %s'
            % cmd_str
        )
        self.job = Job(
            'bash -c "%s"' % cmd_str, env, task_spec, self.datastore._storage_impl
        )
        self.job.submit()

    def wait(self, stdout_location, stderr_location, echo=None):
        def wait_for_launch(job):
            status = job.status
            echo(
                "Task status: %s..." % status,
                "stderr",
                _id=job.id,
            )

        prefix = b"[%s] " % util.to_bytes(self.job.id)
        stdout_tail = get_log_tailer(stdout_location, self.datastore.TYPE)
        stderr_tail = get_log_tailer(stderr_location, self.datastore.TYPE)

        # 1) Loop until the job has started
        wait_for_launch(self.job)

        # 2) Tail logs until the job has finished
        tail_logs(
            prefix=prefix,
            stdout_tail=stdout_tail,
            stderr_tail=stderr_tail,
            echo=echo,
            has_log_updates=lambda: self.job.is_running,
        )

        echo(
            "Task finished with exit code %s." % self.job.result.get("exit_code"),
            "stderr",
            _id=self.job.id,
        )
        if self.job.has_failed:
            raise NvcfException("This could be a transient error. Use @retry to retry.")


class JobStatus(object):
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    SUCCESSFUL = "SUCCESSFUL"
    FAILED = "FAILED"


nvcf_url = "https://api.nvcf.nvidia.com"
submit_endpoint = f"{nvcf_url}/v2/nvcf/pexec/functions"
result_endpoint = f"{nvcf_url}/v2/nvcf/pexec/status"


class Job(object):
    def __init__(self, command, env, task_spec, backend):
        self._payload = {
            "command": command,
            "env": {k: v for k, v in env.items() if v is not None},
        }
        self._result = {}

        conf = init_config()
        if "OBP_AUTH_SERVER" in conf:
            auth_host = conf["OBP_AUTH_SERVER"]
        else:
            auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

        # NOTE: reusing the same auth_host as the one used in NimMetadata,
        # however, user should not need to use nim container to use @nvidia.
        # May want to refactor this to a common endpoint.
        nim_info_url = "https://" + auth_host + "/generate/nim"

        if "METAFLOW_SERVICE_AUTH_KEY" in conf:
            headers = {"x-api-key": conf["METAFLOW_SERVICE_AUTH_KEY"]}
            res = requests.get(nim_info_url, headers=headers)
        else:
            headers = json.loads(os.environ.get("METAFLOW_SERVICE_HEADERS"))
            res = requests.get(nim_info_url, headers=headers)

        res.raise_for_status()
        self._ngc_api_key = res.json()["nvcf"]["api_key"]

        for f in res.json()["nvcf"]["functions"]:
            if f["model_key"] == "metaflow_task_executor":
                self._function_id = f["id"]

        flow_name = task_spec.get("flow_name")
        run_id = task_spec.get("run_id")
        step_name = task_spec.get("step_name")
        task_id = task_spec.get("task_id")
        retry_count = task_spec.get("retry_count")

        heartbeat_prefix = "/".join(
            (flow_name, str(run_id), step_name, str(task_id), str(retry_count))
        )

        ## import is done here to avoid the following warning:
        # RuntimeWarning: 'metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store' found in sys.modules
        # after import of package 'metaflow_extensions.outerbounds.plugins.nvcf', but prior to execution of
        # 'metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store'; this may result in unpredictable behaviour
        from metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store import (
            HeartbeatStore,
        )

        store = HeartbeatStore(
            main_pid=None,
            storage_backend=backend,
        )

        self.heartbeat_thread = threading.Thread(
            target=store.emit_heartbeat,
            args=(
                heartbeat_prefix,
                "nvcf_heartbeats",
            ),
            daemon=True,
        )
        self.heartbeat_thread.start()

    def submit(self):
        try:
            headers = {
                "Authorization": f"Bearer {self._ngc_api_key}",
                "Content-Type": "application/json",
            }
            request_data = json.dumps(self._payload).encode()
            request = Request(
                f"{submit_endpoint}/{self._function_id}",
                data=request_data,
                headers=headers,
            )
            response = urlopen(request)
            self._invocation_id = response.headers.get("NVCF-REQID")
            if response.getcode() == 200:
                data = json.loads(response.read())
                if data["status"].startswith("Oops"):
                    self._status = JobStatus.FAILED
                else:
                    self._status = JobStatus.SUCCESSFUL
                self._result = data
            elif response.getcode() == 202:
                self._status = JobStatus.SUBMITTED
            else:
                self._status = JobStatus.FAILED
            # TODO: Handle 404s nicely
        except (HTTPError, URLError) as e:
            self._state = JobStatus.FAILED
            raise e

    @property
    def status(self):
        if self._status not in [JobStatus.SUCCESSFUL, JobStatus.FAILED]:
            self._poll()
        return self._status

    @property
    def id(self):
        return self._invocation_id

    @property
    def is_running(self):
        return self.status == JobStatus.SUBMITTED

    @property
    def has_failed(self):
        return self.status == JobStatus.FAILED

    @property
    def result(self):
        return self._result

    def _poll(self):
        try:
            headers = {
                "Authorization": f"Bearer {self._ngc_api_key}",
                "Content-Type": "application/json",
            }
            request = Request(
                f"{result_endpoint}/{self._invocation_id}", headers=headers
            )
            response = urlopen(request)
            if response.getcode() == 200:
                data = json.loads(response.read())
                if data["status"].startswith("Oops"):
                    # TODO: Propagate the internal error forward
                    self._status = JobStatus.FAILED
                else:
                    self._status = JobStatus.SUCCESSFUL
                self._result = data
            elif response.getcode() in [400, 500]:
                self._status = JobStatus.FAILED
        except (HTTPError, URLError) as e:
            print(f"Error occurred while polling for result: {e}")



from functools import partial
from metaflow.decorators import FlowDecorator
from metaflow import current
from .nim_manager import NimManager


class NimDecorator(FlowDecorator):
    """
    This decorator is used to run NIM containers in Metaflow tasks as sidecars.

    User code call
    -----------
    @nim(
        models=['meta/llama3-8b-instruct', 'meta/llama3-70b-instruct'],
        backend='managed'
    )

    Valid backend options
    ---------------------
    - 'managed': Outerbounds selects a compute provider based on the model.
    - 🚧 'dataplane': Run in your account.

    Valid model options
    ----------------
        - 'meta/llama3-8b-instruct': 8B parameter model
        - 'meta/llama3-70b-instruct': 70B parameter model
        - Upon request, any model here: https://nvcf.ngc.nvidia.com/functions?filter=nvidia-functions

    Parameters
    ----------
    models: list[NIM]
        List of NIM containers running models in sidecars.
    backend: str
        Compute provider to run the NIM container.
    """

    name = "nim"
    defaults = {
        "models": [],
        "backend": "managed",
    }

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        current._update_env(
            {
                "nim": NimManager(
                    models=self.attributes["models"], backend=self.attributes["backend"]
                )
            }
        )



import os
import time
import json
import requests
from urllib.parse import urlparse
from metaflow.metaflow_config import SERVICE_URL
from metaflow.metaflow_config_funcs import init_config
import sys
import random

NVCF_URL = "https://api.nvcf.nvidia.com"
NVCF_SUBMIT_ENDPOINT = f"{NVCF_URL}/v2/nvcf/pexec/functions"
NVCF_RESULT_ENDPOINT = f"{NVCF_URL}/v2/nvcf/pexec/status"

COMMON_HEADERS = {"accept": "application/json", "Content-Type": "application/json"}
POLL_INTERVAL = 1


class NimMetadata(object):
    def __init__(self):
        self._nvcf_chat_completion_models = []
        self._coreweave_chat_completion_models = []

        conf = init_config()

        if "OBP_AUTH_SERVER" in conf:
            auth_host = conf["OBP_AUTH_SERVER"]
        else:
            auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

        nim_info_url = "https://" + auth_host + "/generate/nim"

        if "METAFLOW_SERVICE_AUTH_KEY" in conf:
            headers = {"x-api-key": conf["METAFLOW_SERVICE_AUTH_KEY"]}
            res = requests.get(nim_info_url, headers=headers)
        else:
            headers = json.loads(os.environ.get("METAFLOW_SERVICE_HEADERS"))
            res = requests.get(nim_info_url, headers=headers)

        res.raise_for_status()
        self._ngc_api_key = res.json()["nvcf"]["api_key"]

        for model in res.json()["nvcf"]["functions"]:
            self._nvcf_chat_completion_models.append(
                {
                    "name": model["model_key"],
                    "function-id": model["id"],
                    "version-id": model["version"],
                }
            )
        for model in res.json()["coreweave"]["containers"]:
            self._coreweave_chat_completion_models.append(
                {"name": model["nim_name"], "ip-address": model["ip_addr"]}
            )

    def get_nvcf_chat_completion_models(self):
        return self._nvcf_chat_completion_models

    def get_coreweave_chat_completion_models(self):
        return self._coreweave_chat_completion_models

    def get_headers_for_nvcf_request(self):
        return {**COMMON_HEADERS, "Authorization": f"Bearer {self._ngc_api_key}"}

    def get_headers_for_coreweave_request(self):
        return COMMON_HEADERS


class NimManager(object):
    def __init__(self, models, backend):
        nim_metadata = NimMetadata()
        if backend == "managed":
            nvcf_models = [
                m["name"] for m in nim_metadata.get_nvcf_chat_completion_models()
            ]
            cw_models = [
                m["name"] for m in nim_metadata.get_coreweave_chat_completion_models()
            ]

            self.models = {}
            for m in models:
                if m in nvcf_models:
                    self.models[m] = NimChatCompletion(
                        model=m, provider="NVCF", nim_metadata=nim_metadata
                    )
                elif m in cw_models:
                    self.models[m] = NimChatCompletion(
                        model=m, provider="CoreWeave", nim_metadata=nim_metadata
                    )
                else:
                    raise ValueError(
                        f"Model {m} not supported by the Outerbounds @nim offering."
                        f"\nYou can choose from these options: {nvcf_models + cw_models}\n\n"
                        "Reach out to Outerbounds if there are other models you'd like supported."
                    )
        else:
            raise ValueError(
                f"Backend {backend} not supported by the Outerbounds @nim offering. Please reach out to Outerbounds."
            )


class NimChatCompletion(object):
    def __init__(
        self,
        model="meta/llama3-8b-instruct",
        provider="CoreWeave",
        nim_metadata=None,
        **kwargs,
    ):
        if nim_metadata is None:
            raise ValueError(
                "NimMetadata object is required to initialize NimChatCompletion object."
            )

        self._nim_metadata = nim_metadata
        self.compute_provider = provider
        self.invocations = []
        self.max_request_retries = int(
            os.environ.get("METAFLOW_EXT_HTTP_MAX_RETRIES", "10")
        )

        if self.compute_provider == "CoreWeave":
            cw_model_names = [
                m["name"]
                for m in self._nim_metadata.get_coreweave_chat_completion_models()
            ]
            self.model = model
            self.ip_address = self._nim_metadata.get_coreweave_chat_completion_models()[
                cw_model_names.index(model)
            ]["ip-address"]
            self.endpoint = f"http://{self.ip_address}:8000/v1/chat/completions"

        elif self.compute_provider == "NVCF":
            nvcf_model_names = [
                m["name"] for m in self._nim_metadata.get_nvcf_chat_completion_models()
            ]
            self.model = model
            self.function_id = self._nim_metadata.get_nvcf_chat_completion_models()[
                nvcf_model_names.index(model)
            ]["function-id"]
            self.version_id = self._nim_metadata.get_nvcf_chat_completion_models()[
                nvcf_model_names.index(model)
            ]["version-id"]

    def __call__(self, **kwargs):

        if self.compute_provider == "CoreWeave":
            request_data = {"model": self.model, **kwargs}
            response = requests.post(
                self.endpoint,
                headers=self._nim_metadata.get_headers_for_coreweave_request(),
                json=request_data,
            )
            response.raise_for_status()
            return response.json()

        elif self.compute_provider == "NVCF":

            request_data = {"model": self.model, **kwargs}
            request_url = f"{NVCF_SUBMIT_ENDPOINT}/{self.function_id}"

            attempts = 0
            while attempts < self.max_request_retries:
                try:
                    attempts += 1
                    response = requests.post(
                        request_url,
                        headers=self._nim_metadata.get_headers_for_nvcf_request(),
                        json=request_data,
                    )
                    response.raise_for_status()
                    if response.status_code == 202:
                        invocation_id = response.headers.get("NVCF-REQID")
                        self.invocations.append(invocation_id)
                    elif response.status_code == 200:
                        return response.json()
                except (
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ) as e:
                    # ConnectionErrors are generally temporary errors like DNS resolution failures,
                    # timeouts etc.
                    print(
                        "received error of type {}. Retrying...".format(type(e)),
                        e,
                        file=sys.stderr,
                    )
                    time.sleep(retry_delay)
                    retry_delay *= 2  # Double the delay for the next attempt
                    retry_delay += random.uniform(0, 1)  # Add jitter
                    retry_delay = min(retry_delay, 10)

            def _poll():
                poll_request_url = f"{NVCF_RESULT_ENDPOINT}/{invocation_id}"
                attempts = 0

                while attempts < self.max_request_retries:
                    try:
                        attempts += 1
                        poll_response = requests.get(
                            poll_request_url,
                            headers=self._nim_metadata.get_headers_for_nvcf_request(),
                        )
                        poll_response.raise_for_status()
                        if poll_response.status_code == 200:
                            return poll_response.json()
                        elif poll_response.status_code == 202:
                            return 202
                        else:
                            raise Exception(
                                f"NVCF returned {poll_response.status_code} status code. Please contact Outerbounds."
                            )
                    except (
                        requests.exceptions.ConnectionError,
                        requests.exceptions.ReadTimeout,
                    ) as e:
                        # ConnectionErrors are generally temporary errors like DNS resolution failures,
                        # timeouts etc.
                        print(
                            "received error of type {}. Retrying...".format(type(e)),
                            e,
                            file=sys.stderr,
                        )
                        time.sleep(retry_delay)
                        retry_delay *= 2  # Double the delay for the next attempt
                        retry_delay += random.uniform(0, 1)  # Add jitter
                        retry_delay = min(retry_delay, 10)

            while True:
                data = _poll()
                if data and data != 202:
                    return data
                time.sleep(POLL_INTERVAL)



import os
from metaflow.decorators import StepDecorator
from metaflow.metadata.metadata import MetaDatum


class InternalFastBakeryDecorator(StepDecorator):
    """
    Internal decorator to support Fast bakery
    """

    name = "fast_bakery_internal"

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        # task_pre_step may run locally if fallback is activated for @catch
        # decorator. In that scenario, we skip collecting Kubernetes execution
        # metadata. A rudimentary way to detect non-local execution is to
        # check for the existence of METAFLOW_KUBERNETES_WORKLOAD environment
        # variable.
        meta = {}
        if "METAFLOW_KUBERNETES_WORKLOAD" in os.environ:
            image = os.environ.get("FASTBAKERY_IMAGE")
            if image:
                meta["fast-bakery-image-name"] = image

        if len(meta) > 0:
            entries = [
                MetaDatum(
                    field=k,
                    value=v,
                    type=k,
                    tags=["attempt_id:{0}".format(retry_count)],
                )
                for k, v in meta.items()
                if v is not None
            ]
            # Register book-keeping metadata for debugging.
            metadata.register_metadata(run_id, step_name, task_id, entries)



import metaflow.metaflow_config_funcs

from metaflow_extensions.outerbounds.remote_config import init_config

# we want to overide OSS Metaflow's initialization behavior with our own to support remote configs
# we're reassigning the METAFLOW_CONFIG variable because all downstream settings rely on it and
# users still have the power to overriden them with environment variables
metaflow.metaflow_config_funcs.METAFLOW_CONFIG = init_config()
metaflow.metaflow_config_funcs.init_config = init_config
